/***********************************************************************************************

      File: 
     Title: APT Automatic Parallelizing Tool
    Author: Nils Schmeisser
   Version: 1.0
  Rev.Date: 20.2.1996

   Comment: 

***********************************************************************************************/

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <apt.h>
#include <dep.h>
#include <fileIO.h>
#include <cppmacro.h>

int FORTRAN_FORMAT_MARK;
int SOURCE,PLATFORM;
int PMODE;
int NMODULES;
int NPROC;
char CFLAGS[1024];
char LFLAGS[1024];
char LIBS[1024];
char **MODULES;
char *MAINMODULE;
ciliststruct *CODEINFO;
blockliststruct *BLOCKS;

char *apt_title    ="APT Automatic Parallelization Tool";
char *apt_version  ="1.0";
char *apt_copyright="(C) Nils Schmeisser, FVTK FZ-Rossendorf, 1996";

const char srcsuffix[4][4]={"c","f","p","asm"};
static int aptmsg=1;

//--------------------------------------------------------------------

blockliststruct *FindBlock(char *module,unsigned long int l1,unsigned long int l2) {
  blockliststruct *p;
  int found=0;

  p=BLOCKS;
  while ((p!=NULL) && (!found)) {
    if (strstr(module,p->module)!=NULL) {
      if (l1==0) {
	if (p->line2==l2) found=1;
      } else {
	if (p->line1==l1) found=1;
      }
    }
    if (!found) p=p->next;
  }
  return p;
}

//--------------------------------------------------------------------

void APTDepCheckerError(char *emsg) {
  printf("*** APT Dependency Checker Error: %s\n",emsg);
}

void APTMessage(char *msg) {
  if (aptmsg)
    printf("--- APT Message: %s\n",msg);
}

void APTError(char *emsg) {
  if (aptmsg)
    printf("*** APT Error: %s\n",emsg);
}

void APTWarning(char *emsg) {
  if (aptmsg)
    printf("!!! APT Warning: %s\n",emsg);
}

//--------------------------------------------------------------------

void InitAPT(char *profile_cfg) {
  FILE *cfgfile;
  char *s,*t;
  int result,i;
  blockliststruct *pl;
  unsigned long int i1,i2;
  
  APTMessage("Initializing Automatic Parallelization Tool");
  s=new(char,1024);
  t=new(char,1024);
  cfgfile=fopen(profile_cfg,"r");
  if (cfgfile!=NULL) {
    result=GetLine(cfgfile,s);
    if (strncmp(s,"APT Configuration File",22)) {
      APTError("Missing 'APT Configuration File' statement at beginning,\nprobably not a configuration file.");
      exit(-1);
    } else if (result==0) {
      APTError("Unexpected end of file.");
      exit(-1);
    }
    result=GetLine(cfgfile,s);
    if (strncmp(s,"Source:",7)) {
      APTError("Missing 'Source:' statement.");
      exit(-1);
    } else if (result==0) {
      APTError("Unexpected end of file.");
      exit(-1);
    }
    sscanf(s,"Source: %s",t);
    if (!(strcmp(t,"C")))
      SOURCE=C_LANGUAGE;
    else if (!(strcmp(t,"FORTRAN")))
      SOURCE=FORTRAN_LANGUAGE;
    else if (!(strcmp(t,"PASCAL")))
      SOURCE=PASCAL_LANGUAGE;
    else if (!(strcmp(t,"ASSEMBLER")))
      SOURCE=ASM_LANGUAGE;
    else {
      sprintf(s,"Unknown source tag %s, possible tags are:\n'C', 'FORTRAN', 'PASCAL' and 'ASSEMBLER'",t);
      APTError(s);
      exit(-1);
    }
    
    result=GetLine(cfgfile,s);
    if (strncmp(s,"Platform:",9)) {
      APTError("Missing 'Platform:' statement.");
      exit(-1);
    } else if (result==0) {
      APTError("Unexpected end of file.");
      exit(-1);
    }
    sscanf(s,"Platform: %s",t);
    if (!(strcmp(t,"UNIX")))
      PLATFORM=UNIX_TARGET;
    else if (!(strcmp(t,"DOS")))
      PLATFORM=DOS_TARGET;
    else if (!(strcmp(t,"OS2")))
      PLATFORM=OS2_TARGET;
    else {
      sprintf(s,"Unknown platform tag %s, possible tags are:\n'UNIX', 'DOS' and 'OS2'",t);
      APTError(s);
      exit(-1);
    }

    result=GetLine(cfgfile,s);
    if (strncmp(s,"NProc:",6)) {
      APTError("Missing 'NProc:' statement.");
      exit(-1);
    } else if (result==0) {
      APTError("Unexpected end of file.");
      exit(-1);
    }
    sscanf(s,"NProc: %d",&NPROC);

    result=GetLine(cfgfile,s);
    if (strncmp(s,"CFlags:",7)) {
      APTError("Missing 'CFlags:' statement.");
      exit(-1);
    } else if (result==0) {
      APTError("Unexpected end of file.");
      exit(-1);
    }
    sscanf(s,"CFlags: %s",CFLAGS);
    
    result=GetLine(cfgfile,s);
    if (strncmp(s,"LFlags:",7)) {
      APTError("Missing 'LFlags:' statement.");
      exit(-1);
    } else if (result==0) {
      APTError("Unexpected end of file.");
      exit(-1);
    }
    sscanf(s,"LFlags: %s",LFLAGS);
    
    result=GetLine(cfgfile,s);
    if (strncmp(s,"Libs:",5)) {
      APTError("Missing 'Libs:' statement.");
      exit(-1);
    } else if (result==0) {
      APTError("Unexpected end of file.");
      exit(-1);
    }
    sscanf(s,"Libs: %s",LIBS);
    
    result=GetLine(cfgfile,s);
    if (strncmp(s,"Modules:",8)) {
      APTError("Missing 'Modules:' statement.");
      exit(-1);
    } else if (result==0) {
      APTError("Unexpected end of file.");
      exit(-1);
    }
    s+=9;
    delete(t);
    t=s;
   
    NMODULES=0;
    while (strlen(t)>0) {
      while (t[0]==' ') t++;
      NMODULES++;
      while ((strlen(t)>0) && (t[0]!=' ')) t++;
    }
   
    MODULES=new(char*,NMODULES);
    for (i=0;i<NMODULES;i++)
      MODULES[i]=new(char,128);
    i=0;
    t=new(char,1024);
    while (strlen(s)>0) {
      while (s[0]==' ') s++;
      sscanf(s,"%s",MODULES[i]);
      while ((strlen(s)>0) && (s[0]!=' ')) s++;
      sprintf(t,"Module #%d is %s",i,MODULES[i]);
      APTMessage(t);
      i++;
    }

    result=GetLine(cfgfile,s);
    if (strncmp(s,"Mode:",5)) {
      APTError("Missing 'Mode:' statement.");
      exit(-1);
    } else if (result==0) {
      APTError("Unexpected end of file.");
      exit(-1);
    }
    sscanf(s,"Mode: %s",t);
    if (!(strcmp(t,"ALL")))
      PMODE=ALL_MODE;
    else if (!(strcmp(t,"LOOP")))
      PMODE=LOOP_MODE;
    else if (!(strcmp(t,"BLOCK")))
      PMODE=BLOCK_MODE;
    else {
      sprintf(s,"Unknown mode tag %s, possible tags are:\n'ALL', 'LOOP' and 'BLOCK'",t);
      APTError(s);
      exit(-1);
    }

    CODEINFO=NULL;
    BLOCKS=NULL;
    if (PMODE==BLOCK_MODE) {
      APTMessage("Block mode, building block list");
      do {
	result=GetLine(cfgfile,s);
	if (strstr(s,"Block: ")!=NULL) {
	  if (BLOCKS==NULL) {
	    BLOCKS=new(blockliststruct,1);
	    pl=BLOCKS;
	  } else {
	    pl->next=new(blockliststruct,1);
	    pl=pl->next;
	  }
	  sscanf(s,"Block: %s %d %d",t,&i1,&i2);
	  pl->module=new(char,strlen(t)+1);
	  strcpy(pl->module,t);
	  pl->line1=i1;
	  pl->line2=i2;
	  pl->next=NULL;
	  sprintf(t,"block in module %s from %d to %d",pl->module,pl->line1,pl->line2);
	  APTMessage(t);
	}
      } while ((strstr(s,"Block: ")!=NULL) && (result!=0));
    } else if (PMODE==LOOP_MODE) {
      APTMessage("Loop mode");
    } else
      APTMessage("All mode");

    fclose(cfgfile);
    
    FORTRAN_FORMAT_MARK=99999;
    MAINMODULE=NULL;
    for (i=0;i<NMODULES;i++) if (MAINMODULE==NULL) {
      switch (SOURCE) {
              case C_LANGUAGE: sprintf(s,"grep \"main(\" %s > /dev/null",MODULES[i]); break;
	case FORTRAN_LANGUAGE: sprintf(s,"grep \"C_APT_MAIN:\" %s > /dev/null",MODULES[i]); break;
      }
      result=system(s);
      if (result==0) {
	sprintf(s,"detected MAIN in %s",MODULES[i]);
	APTMessage(s);
	MAINMODULE=new(char,strlen(MODULES[i])+1);
	strcpy(MAINMODULE,MODULES[i]);
      }
    }
    if (MAINMODULE==NULL) {
      switch (SOURCE) {
        case C_LANGUAGE: APTError("Could not detect main module,\nadd module containing 'int main()'"); 
                         break;
        case FORTRAN_LANGUAGE: 
                         APTError("Could not detect main module,\nadd comment 'C_APT_MAIN' to start of main program"); 
                         break;
      }
      exit(-1);
    }
  } else {
    sprintf(s,"Can't open configuration file %s\n",profile_cfg);
    APTError(s);
  }
}

//--------------------------------------------------------------------

void copyright() {
  printf("%s V %s\n",apt_title,apt_version);
  printf("%s\n",apt_copyright);
  printf("\n");
}

void usage() {
  printf("Usage: apt [ -h | -hcfg | -cfg configfile | -msg ]\n");
  printf("       If no configuration file is given 'apt.cfg'\n");
  printf("       is will be used.\n");
  printf("       '-msg' turns messages off.\n");
  printf("       '-hcfg' gives a detailed help on writing a .cfg file.\n");
}

void usagecfg() {

  usage();
  printf("\n");
  printf("APT Configuration File\n");
  printf("Source: { 'C' | 'FORTRAN' | 'PASCAL' | 'ASSEMBLER' }\n");
  printf("Platform: { 'UNIX' | 'DOS' | 'OS2' }\n");
  printf("Nproc: <# of processors>\n");
  printf("CFlags: [<compiler flags>]\n");
  printf("LFlags: [<linker flags>]\n");
  printf("Libs: [<libraries>]\n");
  printf("Modules: <list of modules to be profiled seperated by a space with NO LINEBREAK>\n");
  printf("Mode: { 'ALL' | 'BLOCK' | 'LOOP' }\n");
  printf("[ Block: <module> <line1> <line2>\n");
  printf(" [<blockdef>] ]\n");
}

//--------------------------------------------------------------------

void CreateMakefile() {
  FILE *mf;
  int i;

  mf=fopen("Makefile.apt","w+");

  fprintf(mf,"# %s V %s\n",apt_title,apt_version);
  fprintf(mf,"# %s\n",apt_copyright);
  fprintf(mf,"#\n");
  switch (SOURCE) {
    case C_LANGUAGE:       fprintf(mf,"COMPILER = cc\n"); break;
    case FORTRAN_LANGUAGE: fprintf(mf,"COMPILER = f77\n"); break;
  }
  fprintf(mf,"CFLAGS   = %s\n",CFLAGS);
  fprintf(mf,"LFLAGS   = %s -L.\n",LFLAGS);
  fprintf(mf,"LIBS     = %s -laptrtm%s\n",LIBS,srcsuffix[SOURCE/2]);
  fprintf(mf,"\n");
  fprintf(mf,"apt.exe:\t");
  for (i=0;i<NMODULES;i++)
    fprintf(mf,"%s.apt.o ",MODULES[i]);
  fprintf(mf,"\n");
  fprintf(mf,"\t$(COMPILER) $(LFLAGS) -o apt.exe \\\n");
  for (i=0;i<NMODULES;i++)
    fprintf(mf,"\t%s.apt.o \\\n",MODULES[i]);
  fprintf(mf,"\t$(LIBS)\n");
  for (i=0;i<NMODULES;i++) {
    fprintf(mf,"\n");
    fprintf(mf,"%s.apt.o:\t%s.apt.%s\n",MODULES[i],MODULES[i],srcsuffix[SOURCE/2]);
    fprintf(mf,"\t$(COMPILER) $(CFLAGS) -c %s.apt.%s\n",MODULES[i],srcsuffix[SOURCE/2]);
  }
  fclose(mf);
}

//--------------------------------------------------------------------

int main(int argc,char **argv) {
  FILE *fin,*fout;
  int i,result;
  char *s,*t,*m,*h;
  int error=0;
  unsigned long int *count,linecount,lch;
  int in_block;
  blockliststruct *bl;
  codeinfostruct *ci;
  char cfgfile[1024];

  copyright();

  strcpy(cfgfile,"apt.cfg");
  if (argc>1) {
    for (i=1;i<argc;i++) {
      if (!(strcmp(argv[i],"-cfg")))
	InitAPT(argv[i+1]);
      if (!(strcmp(argv[i],"-hcfg"))) {
	usagecfg();
	exit(0);
      }
      if (!(strcmp(argv[i],"-h"))) {
	usage();
	exit(0);
      }
      if (!(strcmp(argv[i],"-msg"))) {
	aptmsg=0;
      }
      if (!(strcmp(argv[i],"-cfg"))) {
	strcpy(cfgfile,argv[i+1]);
      }
    }
  }
  InitAPT(cfgfile);

  if (NMODULES>0) {
    s=new(char,1024);
    t=new(char,1024);
    m=new(char,1024);
    InitDepChecker();
    APTMessage("Preparing code information data and profiling hints");
    count=new(unsigned long int,5);
    for (i=0;i<5;i++) count[i]=0;
    for (i=0;i<NMODULES;i++) {
      fin=fopen(MODULES[i],"r");
      if (fin!=NULL) {
	sprintf(s,"%s.apt.%s",MODULES[i],srcsuffix[SOURCE/2]);
	linecount=0;
	fout=fopen(s,"w+");
	sprintf(m,"Profiler: Inserting profiler code to %s -> %s",MODULES[i],s);
	APTMessage(m);
	switch (SOURCE) {
	  case C_LANGUAGE: fprintf(fout,"#include \"aptrtm.h\"\n");
	                   break;
	  case FORTRAN_LANGUAGE: sprintf(s,"grep PROGRAM %s > /dev/null",MODULES[i]);
	                         result=system(s);
	                         if (result==0) {
				   do {
				     result=GetLine(fin,s);
                                     linecount++;
				     fprintf(fout,"%s\n",s);
				   } while (strstr(s,"PROGRAM")==NULL);
				 }
				 fprintf(fout,"      INCLUDE 'APTRTM'\n");
				 break;
	}
	if (PMODE==BLOCK_MODE) {
	  if (BLOCKS!=NULL) {
	    in_block=0;
	    do {
	      do {
		result=GetLine(fin,s);
		linecount++;
		if ((bl=FindBlock(MODULES[i],linecount,0))!=NULL)
		  in_block=1;
		else
		  ci=GenerateCodeInfo(s,MODULES[i],linecount,linecount,count,
				      CODE_SEQUENCE,PROFILING | NO_CODE_INFO,fout);	      
	      } while ((result!=0) && (!in_block));
	      lch=linecount;
	      if ((in_block) && (bl!=NULL)) {
		if (bl->line1==bl->line2)
		  ci=GenerateCodeInfo(s,MODULES[i],lch,linecount,count,
				      CODE_SEQUENCE,PROFILING,fout);	      
		else if ((result!=0) && (in_block)) {
		  do {
		    h=s;
		    result=GetLine(fin,t);
		    linecount++;
		    s=new(char,strlen(h)+1+strlen(t)+1);
		    sprintf(s,"%s\n%s",h,t);
		    delete(h);
		    if (FindBlock(MODULES[i],0,linecount)!=NULL)
		      in_block=0;
		  } while ((result!=0) && (in_block));
		  if ((result!=0) && (!in_block))
		    ci=GenerateCodeInfo(s,MODULES[i],lch,linecount,count,
					CODE_SEQUENCE,PROFILING,fout);	      
		  else {
		    sprintf(t,"File %s truncated, can't find block end",MODULES[i]);
		    APTError(t);
		    error++;
		  }
		}
	      }
	    } while (result!=0);
	  } else
	    error++;
	} else do {
	  result=GetLine(fin,s);
	  if (result!=0) {
	    linecount++;
	    ci=GenerateCodeInfo(s,MODULES[i],linecount,linecount,count,
				CODE_STATEMENT,PROFILING,fout);
	  }
	} while (result!=0);
	fclose(fin);
	fclose(fout);
      } else {
	sprintf(s,"Can't find module %s.",MODULES[i]);
	APTError(s);
	error++;
      }
    }
    APTMessage("Profiler Status Report:");
    sprintf(s,"There are %d errors.",error);
    APTMessage(s);
    if (error==0) {
      APTMessage("now creating Makefile.apt");
      CreateMakefile();
      APTMessage("cleaning up");
      system("rm -f apt.exe *.apt.o > /dev/null");
      APTMessage("now running 'make -f Makefile.apt -k'");
      error=system("make -f Makefile.apt -k");
      if (error==0) {
	sprintf(s,"Now running %s, this may take a while",MAINMODULE);
	APTMessage(s);
	error=system("apt.exe");
	sprintf(s,"Execution returned with status %d",error);
	APTMessage(s);
	if (error==0) {
	  BuildDependencies(CODEINFO);
	  BuildSchedule(CODEINFO,NPROC);
	}
      } else
	APTError("there was an error during execution of 'make',\n-> check for possible syntax errors in YOUR source");
    } else
      APTMessage("Sorry, don't know what to do now");
  } else
    APTError("Nothing to do.");
  return 0;
}




